@@ -3,24 +3,6 @@
  * Copyright (C) 2007 Red Hat, Inc.
  */
 
-#include <ctype.h>
-#include <fcntl.h>
-#include <errno.h>
-#include <signal.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <sys/stat.h>
-#include <unistd.h>
-
-#include <glib.h>
-#include <libsoup/soup-address.h>
-#include <libsoup/soup-message.h>
-#include <libsoup/soup-misc.h>
-#include <libsoup/soup-server.h>
-#include <libsoup/soup-session-async.h>
-#include <libsoup/soup-session-sync.h>
-
 #include "test-utils.h"
 
 static char *base_uri;
